// Copyright 2024 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import 'dart:convert';

import 'package:google_generative_ai/google_generative_ai.dart';
import 'package:google_generative_ai/src/api.dart';
import 'package:test/test.dart';

import 'utils/matchers.dart';

void main() {
  group('throws errors for invalid GenerateContentResponse', () {
    test('with empty content', () {
      final response = '''
{
  "candidates": [
    {
      "content": {},
      "index": 0
    }
  ],
  "promptFeedback": {
    "safetyRatings": [
      {
        "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
        "probability": "NEGLIGIBLE"
      },
      {
        "category": "HARM_CATEGORY_HATE_SPEECH",
        "probability": "NEGLIGIBLE"
      },
      {
        "category": "HARM_CATEGORY_HARASSMENT",
        "probability": "NEGLIGIBLE"
      },
      {
        "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
        "probability": "NEGLIGIBLE"
      }
    ]
  }
}
''';
      final decoded = jsonDecode(response) as Object;
      expect(
          () => parseGenerateContentResponse(decoded),
          throwsA(isA<FormatException>().having((e) => e.message, 'message',
              startsWith('Unhandled Content format'))));
    });

    test('with a blocked prompt', () {
      final response = '''
{
  "promptFeedback": {
    "blockReason": "SAFETY",
    "safetyRatings": [
      {
        "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
        "probability": "NEGLIGIBLE"
      },
      {
        "category": "HARM_CATEGORY_HATE_SPEECH",
        "probability": "HIGH"
      },
      {
        "category": "HARM_CATEGORY_HARASSMENT",
        "probability": "NEGLIGIBLE"
      },
      {
        "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
        "probability": "NEGLIGIBLE"
      }
    ]
  }
}
''';
      final decoded = jsonDecode(response) as Object;
      final generateContentResponse = parseGenerateContentResponse(decoded);
      expect(
          generateContentResponse,
          matchesGenerateContentResponse(GenerateContentResponse(
              [],
              PromptFeedback(BlockReason.safety, null, [
                SafetyRating(
                    HarmCategory.sexuallyExplicit, HarmProbability.negligible),
                SafetyRating(HarmCategory.hateSpeech, HarmProbability.high),
                SafetyRating(
                    HarmCategory.harassment, HarmProbability.negligible),
                SafetyRating(
                    HarmCategory.dangerousContent, HarmProbability.negligible),
              ]))));
      expect(
          () => generateContentResponse.text,
          throwsA(isA<GenerativeAIException>().having((e) => e.message,
              'message', startsWith('Response was blocked due to safety'))));
    });
  });

  group('parses successful GenerateContentResponse', () {
    test('with a basic reply', () async {
      final response = '''
{
  "candidates": [
    {
      "content": {
        "parts": [
          {
            "text": "Mountain View, California, United States"
          }
        ],
        "role": "model"
      },
      "finishReason": "STOP",
      "index": 0,
      "safetyRatings": [
        {
          "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
          "probability": "NEGLIGIBLE"
        },
        {
          "category": "HARM_CATEGORY_HATE_SPEECH",
          "probability": "NEGLIGIBLE"
        },
        {
          "category": "HARM_CATEGORY_HARASSMENT",
          "probability": "NEGLIGIBLE"
        },
        {
          "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
          "probability": "NEGLIGIBLE"
        }
      ]
    }
  ],
  "promptFeedback": {
    "safetyRatings": [
      {
        "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
        "probability": "NEGLIGIBLE"
      },
      {
        "category": "HARM_CATEGORY_HATE_SPEECH",
        "probability": "NEGLIGIBLE"
      },
      {
        "category": "HARM_CATEGORY_HARASSMENT",
        "probability": "NEGLIGIBLE"
      },
      {
        "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
        "probability": "NEGLIGIBLE"
      }
    ]
  }
}
''';
      final decoded = jsonDecode(response) as Object;
      final generateContentResponse = parseGenerateContentResponse(decoded);
      expect(
          generateContentResponse,
          matchesGenerateContentResponse(GenerateContentResponse(
              [
                Candidate(
                    Content.model(
                        [TextPart('Mountain View, California, United States')]),
                    [
                      SafetyRating(HarmCategory.sexuallyExplicit,
                          HarmProbability.negligible),
                      SafetyRating(
                          HarmCategory.hateSpeech, HarmProbability.negligible),
                      SafetyRating(
                          HarmCategory.harassment, HarmProbability.negligible),
                      SafetyRating(HarmCategory.dangerousContent,
                          HarmProbability.negligible),
                    ],
                    null,
                    FinishReason.stop,
                    null),
              ],
              PromptFeedback(null, null, [
                SafetyRating(
                    HarmCategory.sexuallyExplicit, HarmProbability.negligible),
                SafetyRating(
                    HarmCategory.hateSpeech, HarmProbability.negligible),
                SafetyRating(
                    HarmCategory.harassment, HarmProbability.negligible),
                SafetyRating(
                    HarmCategory.dangerousContent, HarmProbability.negligible),
              ]))));
    });

    test('with a citation', () async {
      final response = '''
{
  "candidates": [
    {
      "content": {
        "parts": [
          {
            "text": "placeholder"
          }
        ],
        "role": "model"
      },
      "finishReason": "STOP",
      "index": 0,
      "safetyRatings": [
        {
          "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
          "probability": "NEGLIGIBLE"
        },
        {
          "category": "HARM_CATEGORY_HATE_SPEECH",
          "probability": "NEGLIGIBLE"
        },
        {
          "category": "HARM_CATEGORY_HARASSMENT",
          "probability": "NEGLIGIBLE"
        },
        {
          "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
          "probability": "NEGLIGIBLE"
        }
      ],
      "citationMetadata": {
        "citationSources": [
          {
            "startIndex": 574,
            "endIndex": 705,
            "uri": "https://example.com/",
            "license": ""
          },
          {
            "startIndex": 899,
            "endIndex": 1026,
            "uri": "https://example.com/",
            "license": ""
          }
        ]
      }
    }
  ],
  "promptFeedback": {
    "safetyRatings": [
      {
        "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
        "probability": "NEGLIGIBLE"
      },
      {
        "category": "HARM_CATEGORY_HATE_SPEECH",
        "probability": "NEGLIGIBLE"
      },
      {
        "category": "HARM_CATEGORY_HARASSMENT",
        "probability": "NEGLIGIBLE"
      },
      {
        "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
        "probability": "NEGLIGIBLE"
      }
    ]
  }
}
''';
      final decoded = jsonDecode(response) as Object;
      final generateContentResponse = parseGenerateContentResponse(decoded);
      expect(
          generateContentResponse,
          matchesGenerateContentResponse(GenerateContentResponse(
              [
                Candidate(
                    Content.model([TextPart('placeholder')]),
                    [
                      SafetyRating(HarmCategory.sexuallyExplicit,
                          HarmProbability.negligible),
                      SafetyRating(
                          HarmCategory.hateSpeech, HarmProbability.negligible),
                      SafetyRating(
                          HarmCategory.harassment, HarmProbability.negligible),
                      SafetyRating(HarmCategory.dangerousContent,
                          HarmProbability.negligible),
                    ],
                    CitationMetadata([
                      CitationSource(
                          574, 705, Uri.https('example.com', ''), ''),
                      CitationSource(
                          899, 1026, Uri.https('example.com', ''), ''),
                    ]),
                    FinishReason.stop,
                    null),
              ],
              PromptFeedback(null, null, [
                SafetyRating(
                    HarmCategory.sexuallyExplicit, HarmProbability.negligible),
                SafetyRating(
                    HarmCategory.hateSpeech, HarmProbability.negligible),
                SafetyRating(
                    HarmCategory.harassment, HarmProbability.negligible),
                SafetyRating(
                    HarmCategory.dangerousContent, HarmProbability.negligible),
              ]))));
    });
  });
}
